function transition_entity(ent, transition)
        for i=0,ent.numSubEntities-1 do
                local mname = ent:getMaterialName(i)
                if mname:sub(-1) == "'" then mname = mname:sub(1,#mname-1) end
                if transition == 0 then
                        --ent.visible = true
                        ent:setMaterial(i,mname)
                        --elseif transition == 1 then
                        --        ent.visible = false
                else
                        --ent.visible = true
                        ent:setMaterial(i,mname.."'")
                end
        end
end

local function get_rot (x)
        if x == nil then return Q_ID end
        if getmetatable(x) == getmetatable(V_ID) then 
                return Quat(x.z,V_UP) * Quat(x.y,V_NORTH) * Quat(x.x,V_EAST)
        end
        return x
end

BaseClass = {
        renderingDistance=400;
        init = function (persistent)
                persistent:hintAdvancePrepare("Mesh",persistent.gfxMesh or persistent.className..".mesh")
                if persistent.textures == nil then return end
                for _,v in ipairs(persistent.textures) do
                    persistent:hintAdvancePrepare("Texture", v)
                end
        end;
        activate=function (class,persistent,gfxRoot)
                --echo("Activating: "..persistent.name.." ("..persistent.className..")")
                local self = {}
                self.persistent = persistent
                self.class = class
                self.renderingDistance = class.renderingDistance
                self.gfx = gfxRoot:createChild()
                local gfxMesh = persistent.gfxMesh or class.gfxMesh or persistent.className..".mesh"
                self.entity = gfx.scnmgr:createMovableObject(persistent.name,"Entity",{mesh=gfxMesh})
                self.entity.renderingDistance = self.renderingDistance
                self.entity:setCustomParameterAll(0,1,persistent.near and 1 or 0,0,0)
                -- what we should do:
                -- 1) render everything that is not being LOD-faded
                -- 2) render the alpha parts of the lod fades without depth write
                -- 3) render the non-alpha parts with depth write
                -- not 100% sure it will work though
                --if persistent.near then
                --        self.entity.renderQueueGroup = 40;
                --end
                self.entity.castShadows = persistent.castShadows
                if persistent.scale ~= nil then
                        self.gfx.scale = persistent.scale
                end
                self.gfx.position = persistent.spawnPos
                self.gfx.orientation = get_rot(persistent.rot)
                self.gfx:attachObject(self.entity)
                --persistent:updateSphere(persistent.pos.x, persistent.pos.y, persistent.pos.z, self.renderingDistance)
                return self
        end;
        setFade=function(self, fade, transition)
                --local persistent = self.persistent
                --print(persistent.name,fade,transition)
                local ent = self.entity
                if transition then
                        -- slow path
                        transition_entity(ent, transition)
                end
                ent:setCustomParameterAll(0,0,fade)
                ent.visible = fade > 0.001
        end;
        deactivate=function(self)
                --echo("Deactivating: "..self.persistent.name.." ("..self.persistent.className..")")
                local pos = self.persistent.spawnPos
                self.persistent:updateSphere(pos.x,pos.y,pos.z,self.renderingDistance)
                self.gfx = safe_destroy(self.gfx)
                self.entity = safe_destroy(self.entity)
                return self.persistent.norespawn
        end
}

ColClass = extends (BaseClass) {
        activate=function (class,persistent,gfxRoot,physicsWorld)
                local self = BaseClass.activate(class,persistent,gfxRoot,physicsWorld)
                local colMesh = class.colMesh or persistent.className..".tcol"
                --echo("adding: "..tostring(persistent).." with "..colMesh)
                self.world = physicsWorld
                self.body = physicsWorld:addRigidBody(
                        colMesh,
                        persistent.spawnPos,
                        get_rot(persistent.rot)
                )
                self.body.owner = persistent;
                if persistent.floating then
                        self.body:deactivate()
                end
                -- this causes an alloc so do them here where the code is cold
                self.body.updateCallback = function (px,py,pz,rw,rx,ry,rz)
                        local node = self.gfx
                        -- can be nil
                        if node ~= nil and alive(node) then
                                node:setPosition(px,py,pz)
                                node:setOrientation(rw,rx,ry,rz)
                        end
                        persistent:updateSphere(px,py,pz,self.renderingDistance)
                end
                return self
        end;
        deactivate=function(self)
                self.body = safe_destroy(self.body)
                return BaseClass.deactivate(self)
        end;

        getSpeed = function (self)
                if not self.activated then error("Not activated: "..self.persistent.name) end
                local rb = self.instance.body
                return #rb.linearVelocity
        end;
        setSpeed = function (self,speed)
                if not self.activated then error("Not activated: "..self.persistent.name) end
                local rb = self.instance.body
                rb.linearVelocity = rb.linearVelocity.normalised * speed
                rb:activate();
        end;
        boost = function (self,speed)
                speed = speed or 1
                self:setSpeed(self:getSpeed()+speed)
        end;
}

-- vim: shiftwidth=8:tabstop=8:expandtab
