print("Loading gfx.lua")

GFX = GFX or { }

function GFX:showStats(b)
        self.luaCounter.visible = b
        self.fpsCounter.visible = b
        self.triangleBatchCounter.visible = b
        self.memCounter.visible = b
end

function GFX:setShadow(b0,b1,b2,b3,oaf0,oaf1,oaf2,padding)
        b0 = b0 or 0.2
        b1 = b1 or 20
        b2 = b2 or 50
        b3 = b3 or 200
        oaf0 = oaf0 or 3
        oaf1 = oaf1 or 1
        oaf2 = oaf2 or 1
        padding = padding or 0.6
        self.scnmgr:setShadowCameraSetup("PSSM",{
                splits=3,
                padding=padding,
                boundary0=b0,
                boundary1=b1,
                boundary2=b2,
                boundary3=b3,
                optimalAdjustFactor0=oaf0,
                optimalAdjustFactor1=oaf1,
                optimalAdjustFactor2=oaf2,
        })
end

function GFX.new()
        local self = { }

        self.scnmgr = SceneManager("OctreeSceneManager","scnmgr")

        --self.scnmgr.shadowTextureCasterMaterial = "system/depth";
        self.scnmgr:setShadowTextureCountPerLightType("DIRECTIONAL",3)
        self.scnmgr:setShadowTextureSettings(1024,3,"PF_FLOAT32_R");
        --self.scnmgr:setShadowTextureSettings(2048,1,"PF_FLOAT32_R");
        --self.scnmgr:setShadowTextureSettings(2048,1,"PF_L16");
        --self.scnmgr:setShadowTextureSettings(2048,1,"PF_L8");
        self.scnmgr.shadowTechnique = "TEXTURE_MODULATIVE_INTEGRATED"
        --self.scnmgr.shadowCameraSetup = "FOCUSED"
        --self.scnmgr.shadowCameraSetup = "LISPSM" -- better, but more volatile
        --self.scnmgr.shadowFarDistance = 100
        -- had to comment this out because we have surfaces without normals
        -- on the other hand, some trees have 2sided polys so this is redundant
        self.scnmgr.shadowCasterRenderBackFaces = false

        self.scnmgr:setSkyBox(true,"system/SkyBox",100,true,Q_UP)
        self.scnmgr:setFog("EXP2",0.7137,0.5333,0.4,0.0015,0,0)

        self.scnmgr:setAmbientLight(0.2,0.2,0.2)
        self.sun = self.scnmgr:createLight("Sun")
        self.sun.type = "DIRECTIONAL"
        self.sun.direction = Vector3(.1,.7,-1).normalised
        self.sun:setDiffuseColour(.8,.75,.7)
        self.sun:setSpecularColour(1,.85,.7)


        streamer.gfx = self.scnmgr.root

        self.mainWin = get_main_win()

        self.camPos = self.scnmgr.root:createChild("CamPos")
        self.camYaw = self.camPos:createChild("CamYaw") 
        self.camPitch = self.camYaw:createChild("CamPitch") 
        self.camRoll = self.camPitch:createChild("CamRoll") 

        self.camChase = self.scnmgr.root:createChild("CamChase")
        self.camChase.inheritOrientation = false
        self.camChaseYaw = self.camChase:createChild("CamChaseYaw")
        self.camChasePitch = self.camChaseYaw:createChild("CamChasePitch")
        self.camChaseRoll = self.camChasePitch:createChild("CamChaseRoll")
        self.camChaseBoom = self.camChaseRoll:createChild("CamChaseBoom")
        self.camChaseBoom.position = V_FORWARDS * 8

        self.queueAggressiveness = function () return 1 + 0.5*get_in_queue_size() end


        -- OVERLAY STUFF

        self.crosshair = get_hud_root():addChild("Pane")
        self.crosshair.material = "system/CrossHair"
        self.crosshair.resize = function (pw,ph) return pw/2-2,ph/2-2,4,4 end
       
        local opt = {font = "misc.fixed", charHeight = 13}
        self.fpsCounter = get_hud_root():addChild("ShadowText",opt)
        self.fpsCounter.resize = Hud.RIGHT(-52-1)

        self.streamerCounter = get_hud_root():addChild("ShadowText",opt)
        self.streamerCounter.resize = Hud.RIGHT(-39-1)

        self.luaCounter = get_hud_root():addChild("ShadowText",opt)
        self.luaCounter.resize = Hud.RIGHT(-26-1)

        self.memCounter = get_hud_root():addChild("ShadowText",opt)
        self.memCounter.resize = Hud.RIGHT(-13-1)

        self.triangleBatchCounter = get_hud_root():addChild("ShadowText",opt)
        self.triangleBatchCounter.resize = Hud.RIGHT(-1)


        self.frameTimeMin = RunningMin.new(20)
        self.frameTimeAverage = RunningAverage.new(20)
        self.frameTimeMax = RunningMax.new(20)
        self.frameTimer = Timer()

        local function frameCallback()

                gfx.scnmgr:updateSceneGraph()
                local pos = gfx.cam.derivedPosition
                streamer:centre(pos.x,pos.y,pos.z)

                local _,unacct_allocs = get_alloc_stats()
                reset_alloc_stats()
                render() -- render a frame (takes a while!)
                local count, gfx_allocs = get_alloc_stats()
                reset_alloc_stats()

                give_queue_allowance(self:queueAggressiveness())

                local last_time = self.frameTimer.us/1000000
                self.frameTimeMin:add(last_time)
                self.frameTimeAverage:add(last_time)
                self.frameTimeMax:add(last_time)
                self.frameTimer:reset()

                if self.fpsCounter and self.fpsCounter.visible then
                        local min_fps = 1/nonzero(self.frameTimeMax:calc())
                        local avg_fps = 1/nonzero(self.frameTimeAverage:calc())
                        local max_fps = 1/nonzero(self.frameTimeMin:calc())
                        self.fpsCounter.text = string.format("FPS min / avg / max: %3.0f / %3.0f / %3.0f",min_fps,avg_fps,max_fps)
                        local green = clamp(min_fps/75,0,1)
                        local red = clamp((75-min_fps)/75,0,1)
                        self.fpsCounter:setColourTop(red,green,0,1)
                        self.fpsCounter:setColourBottom(red,green,0,1)
                end
                if self.streamerCounter and self.streamerCounter.visible then
                        self.streamerCounter.text = string.format(
                                "Classes / Objects / Activated: %6.0f / %6.0f / %6.0f",
                                streamer.numClasses, streamer.numObjects, streamer.numActivated)
                end
                if self.memCounter and self.memCounter.visible then
                        self.memCounter.text = string.format(
                                "Mesh | Tex | Max || QI | QO: %dMB | %dMB | %dMB || %d | %d",
                                get_mesh_usage()/1024/1024,
                                get_texture_usage()/1024/1024,
                                (get_mesh_budget()+get_texture_budget())/1024/1024,
                                get_in_queue_size(),
                                get_out_queue_size());
                end
                if self.triangleBatchCounter and self.triangleBatchCounter.visible then
                        self.triangleBatchCounter.text = string.format(
                                "Tris | Batches: %d / %d",self.mainWin.triangles,self.mainWin.batches)
                end
                if self.luaCounter and self.luaCounter.visible then
                        self.luaCounter.text = string.format(
                                "Lua: %.1fMB (%dK) (G+%d) (P+%d) (U+%d)",
                                collectgarbage("count")/1024,
                                count/1000,
                                gfx_allocs,
                                physics and physics.allocs or 0,
                                unacct_allocs)
                end
        end

        main.frameCallbacks:insert("GFX.frameCallback",frameCallback)

        make_instance(self,GFX,function (k)
                if k=="FOV" then
                        return true, self.cam.FOV
                end
        end,function (k,v)
                if k=="FOV" then
                        self.cam.FOV = v
                        return true
                end
        end)
       
        self:recreateCam("MonoCamera")

        self:setShadow()

        return self

end

function GFX:orient(dir) -- direction as vector
        local bearing = dir * Vector3(1,1,0) -- eliminate z component
        self.camYaw.orientation = V_NORTH:getRotationTo(bearing)
        local pitch = V_NORTH:getRotationTo(self.camYaw.orientation.inverse * dir)
        self.camPitch.orientation = pitch
        self.camRoll.orientation = Q_ID
end

function GFX:warp(pos, orientation)
        if pos ~= nil then
                self.camPos.position = pos;
        end
        if orientation ~= nil then
                local dir = V_NORTH * orientation
                self:orient(dir)
        end
end

function GFX:setFog(amount)
        local _, r, g, b, _, _, _ = self.scnmgr:getFog()
        self.scnmgr:setFog("EXP2",r,g,b,amount,0,0)
end

function GFX:retWarp()
        local pos = self.cam.derivedPosition
        local orientation = self.cam.derivedOrientation
        return "gfx:warp("..tostring(pos)..","..tostring(orientation)..")"
end

function GFX:yaw()
        local pointing = gfx.cam.derivedOrientation * V_FORWARDS
        pointing = pointing * Vector3(1,1,0) -- eliminate z component
        return V_NORTH:getRotationTo(pointing)
end

function GFX:yawAngle()
        local quat = gfx:yaw()
        local angle = math.deg(2*math.acos(quat.w))
        if (gfx.cam.derivedOrientation * V_FORWARDS).x < 0 then angle = 360-angle end
        return angle
end

function GFX:setDebugShadow(index, on)
        if self.debugShadowPane == nil then self.debugShadowPane = {} end
        if on then
                if self.debugShadowPane[index]==nil then
                        self.debugShadowPane[index] = get_hud_root():addChild("Pane")
                        self.debugShadowPane[index].material = "DebugShadow"..index
                        self.debugShadowPane[index]:setUV(0,0,1,1)
                        self.debugShadowPane[index].resize = Hud.LEFT(-1-128*index,128,128)
                end
                self.debugShadowPane[index].material:setTextureName(0,0,0,self.scnmgr:getShadowTexture(index).name)
        elseif not on and self.debugShadowPane[index]~=nil then
                get_hud_root():removeChild(self.debugShadowPane[index])
                self.debugShadowPane[index] = nil
        end
end

function GFX:destroy()
        streamer:deactivateAll()
        if self.cam then self.cam:destroy() end
        self.fpsCounter:destroy()
        self.streamerCounter:destroy()
        self.triangleBatchCounter:destroy()
        self.memCounter:destroy()
        self.luaCounter:destroy()
        self.scnmgr:destroy()
        main.frameCallbacks:removeByName("GFX.frameCallback")
end

function GFX:recreateCam(type)
        if self == nil then
                error("Self is nil")
        end

        if not _G[type] then
                error("No such camera: "..type)
                return
        end
        local parent = self.camRoll
        local FOV = 60
        local nearClipDistance = 0.3
        local autoAspectRatio = true
        local useRenderingDistance = false
        if self.cam then
                parent = self.cam.parentSceneNode
                FOV = self.cam.FOV
                nearClipDistance = self.cam.nearClipDistance
                autoAspectRatio = self.cam.autoAspectRatio
                useRenderingDistance = self.cam.useRenderingDistance
                self.cam:destroy()
        end
        self.cam = _G[type].new(self.scnmgr,parent,self.mainWin,1,0,0,1,1)
        self.cam.FOV = FOV
        self.cam.nearClipDistance = nearClipDistance
        self.cam.autoAspectRatio = autoAspectRatio
        self.cam.useRenderingDistance = useRenderingDistance
end


if gfx ~= nil then
        gfx:destroy()
end
gfx = GFX.new()

sm = gfx.scnmgr

-- vim: shiftwidth=8:tabstop=8:expandtab
