
local function process_tex_name (mname, tname)
        local substring = tname:sub(1,2)
        if substring == "./" then
                return (string.gsub(mname, "[^/]+$", tname:sub(3)))
        end
        if substring == "//" then
                return mname.."."..tname:sub(3)
        end
        return tname
end

local function get_colour (v, r, g, b, a)
        if v==nil then return r,g,b,a end
        if type(v)=="number" then
                r = v % 256
                v = (v - r)/256
                g = v % 256
                v = (v - g)/256
                b = v % 256
                v = (v - b)/256
                a = v % 256
                return r/255, g/255, b/255, a/255
        else
                return v[1] or r, v[2] or g, v[3] or b, v[4] or a
        end
end


local depth = get_material("common/depth")

local function shader (v, dm, nm, sm)
        if sm then 
                if v then
                        if dm then
                                if nm then
                                        return "VTNS_f", "VTNS_v"
                                else
                                        return "VTnS_f", "VTnS_v"
                                end
                        else
                                if nm then
                                        return "VtNS_f", "VtNS_v"
                                else
                                        return "VtnS_f", "VtnS_v"
                                end
                        end
                else
                        if dm then
                                if nm then
                                        return "vTNS_f", "vTNS_v"
                                else
                                        return "vTnS_f", "vTnS_v"
                                end
                        else
                                if nm then
                                        return "vtNS_f", "vtNS_v"
                                else
                                        return "vtnS_f", "vtnS_v"
                                end
                        end
                end
        else
                if v then
                        if dm then
                                if nm then
                                        return "VTNs_f", "VTNs_v"
                                else
                                        return "VTns_f", "VTns_v"
                                end
                        else
                                if nm then
                                        return "VtNs_f", "VtNs_v"
                                else
                                        return "Vtns_f", "Vtns_v"
                                end
                        end
                else
                        if dm then
                                if nm then
                                        return "vTNs_f", "vTNs_v"
                                else
                                        return "vTns_f", "vTns_v"
                                end
                        else
                                if nm then
                                        return "vtNs_f", "vtNs_v"
                                else
                                        return "vtns_f", "vtns_v"
                                end
                        end
                end
        end
end

local d3d9 = get_rendersystem() == "Direct3D9 Rendering Subsystem"
local lang = d3d9 and "hlsl" or "glsl"

local function make_programs (vertex_colours, diffuse_map, normal_map, spec_map)
        local name = (vertex_colours and "V" or "v")..(diffuse_map and "T" or "t")..(normal_map and "N" or "n")..(spec_map and "S" or "s")
        local defines = "USE_VERTEX_COLOURS="..(vertex_colours and "1" or "0")
                      ..";USE_DIFFUSE_MAP="..(diffuse_map and "1" or "0")
                      ..";USE_NORMAL_MAP="..(normal_map and "1" or "0")
                      ..";USE_SPECULAR_MAP="..(spec_map and "1" or "0")
        if d3d9 then
                -- for some reason defining GRIT_GLSL didn't work -- seemed macro was not visible
                -- maybe something to do with #version
                defines = defines .. ";GRIT_HLSL=1"
        end

        local vp = get_gpuprog(name.."_v") or make_gpuprog(name.."_v",lang,"VERTEX")
        vp.target = "vs_3_0"
        vp.defines = defines..";VERTEX_PROGRAM=main"
        vp.sourceFile = "system/uber.shader"
        vp.entryPoint = "main"
        vp:setAutoConstantInt("world","WORLD_MATRIX")
        vp:setAutoConstantInt("view_proj","VIEWPROJ_MATRIX")
        for i=1,3 do
                vp:setAutoConstantInt("shadow_world_view_proj"..i,"TEXTURE_WORLDVIEWPROJ_MATRIX",i-1)
        end
        vp:setAutoConstantInt("camera_pos","CAMERA_POSITION_OBJECT_SPACE")
        vp:setAutoConstantInt("light_pos_os","LIGHT_POSITION_OBJECT_SPACE",0)
        vp:setAutoConstantInt("light_pos_ws","LIGHT_POSITION")
        vp:setAutoConstantInt("light_ambient","DERIVED_SCENE_COLOUR")
        vp:setAutoConstantInt("surf_ambient","SURFACE_AMBIENT_COLOUR")

        local fp = get_gpuprog(name.."_f") or make_gpuprog(name.."_f",lang,"FRAGMENT")
        fp.target = "ps_3_0"
        fp.defines = defines..";FRAGMENT_PROGRAM=main"
        fp.sourceFile = "system/uber.shader"
        fp.entryPoint = "main"
        if not d3d9 then
                local tunit = 0
                if diffuse_map then
                        fp:setConstantInt("diffuse_map", tunit);
                        tunit = tunit+1
                end
                if normal_map then
                        fp:setConstantInt("normal_map", tunit);
                        tunit = tunit+1
                end
                if spec_map then
                        fp:setConstantInt("spec_map", tunit);
                        tunit = tunit+1
                end
                for i=1,3 do
                        fp:setConstantInt("shadow_map"..i, tunit);
                        tunit = tunit+1
                end
        end
        fp:setAutoConstantInt("the_fog_params", "FOG_PARAMS")
        fp:setAutoConstantInt("the_fog_colour", "FOG_COLOUR")
        fp:setAutoConstantInt("visibility", "CUSTOM", 0)
        fp:setConstantFloat("alpha_rej", 0)
        fp:setAutoConstantInt("light_diffuse", "DERIVED_LIGHT_DIFFUSE_COLOUR", 0)
        fp:setAutoConstantInt("light_specular", "DERIVED_LIGHT_SPECULAR_COLOUR", 0)
        fp:setAutoConstantInt("surf_diffuse", "SURFACE_DIFFUSE_COLOUR")
        fp:setAutoConstantInt("surf_specular", "SURFACE_SPECULAR_COLOUR")
        if not d3d9 then
                fp:setAutoConstantInt("surf_shininess", "SURFACE_SHININESS")
        end

        fp:reload()
        vp:reload()
        return vp, fp
end

local function material_internal (name, original, fade, tab)

        local lighting = not (tab.normals == false)

        local shf, shv = shader(not lighting, tab.diffuseMap, tab.normalMap, tab.specularMap)

        local mat = get_material(name)
        if mat==nil then
                mat = Material(name)
        else
                mat:removeAllTechniques()
                mat:createTechnique()
                mat:createPass(0)
        end


        mat:setShadowCasterMaterial(0,depth)

        mat:setFragmentProgram(0,0,shf)
        mat:setVertexProgram(0,0,shv)

        mat:setFragmentConstantFloat(0,0,"alpha_rej", tab.alphaReject or 63/255)

        if tab.backfaces then
                mat:setCull(0,0,false)
                mat:setManualCull(0,0,false)
        end     


        -- diffuse map
        local tex_index = 0
        if tab.diffuseMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.diffuseMap))
                mat:setTextureAnisotropy(0,0,tex_index, 4)
                mat:setTextureFiltering(0,0,tex_index,"ANISOTROPIC","ANISOTROPIC","LINEAR")
                tex_index = tex_index + 1
        end

       -- normal map
        if tab.normalMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.normalMap))
                mat:setTextureAnisotropy(0,0,tex_index, 4)
                mat:setTextureFiltering(0,0,tex_index,"ANISOTROPIC","ANISOTROPIC","LINEAR")
                tex_index = tex_index + 1
        end

       -- spec map
        if tab.specularMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.specularMap))
                mat:setTextureAnisotropy(0,0,tex_index, 4)
                mat:setTextureFiltering(0,0,tex_index,"ANISOTROPIC","ANISOTROPIC","LINEAR")
                tex_index = tex_index + 1
        end

        -- diffuse colour
        local r,g,b,a = get_colour(tab.diffuseColour,1,1,1,1)
        mat:setAmbient(0,0, r,g,b)
        if lighting then
                mat:setDiffuse(0,0, r,g,b,a)
        else
                mat:setDiffuse(0,0, 0,0,0,a)
        end

        -- specular colour
        if lighting then
                r,g,b = get_colour(tab.specularColour,0,0,0)
                mat:setSpecular(0,0, r,g,b,1,tab.specularity or 10)
        else
                mat:setSpecular(0,0, 0,0,0,0,0)
        end

        -- alpha stuff
        if fade or tab.alpha then
                mat:setSceneBlending(0,0,"SRC_ALPHA","ONE_MINUS_SRC_ALPHA")
        end
        if tab.alpha then
                mat:setDepthWriteEnabled(0,0,false)
        end
        if fade or tab.zsort then
                mat:setTransparentSortingForced(0,0,true)
        end

        -- shadow maps
        for i=1,3 do
                mat:createTextureUnitState(0,0)
                mat:setIsShadow(0,0,tex_index, true)
                mat:setTextureAnisotropy(0,0,tex_index, 4)
                mat:setTextureFiltering(0,0,tex_index,"ANISOTROPIC","ANISOTROPIC","LINEAR")
                tex_index = tex_index + 1
        end
end

-- curry to avoid parentheses
function material(name) return function(tab)
        tab = tab or {}
        material_internal(name, name, false, tab)
        material_internal(name.."'", name, true, tab)
end end


make_programs(false,true,true,false)
make_programs(false,true,true,true)
make_programs(false,true,false,false)
make_programs(false,true,false,true)
make_programs(false,false,true,false)
make_programs(false,false,true,true)
make_programs(false,false,false,false)
make_programs(false,false,false,true)
make_programs(true,true,true,false)
make_programs(true,true,true,true)
make_programs(true,true,false,false)
make_programs(true,true,false,true)
make_programs(true,false,true,false)
make_programs(true,false,true,true)
make_programs(true,false,false,false)
make_programs(true,false,false,true)
-- vim: ts=8:sw=8:et:tw=100
