print("Loading physics.lua")

Physics = Physics or { }

function Physics.new()
        local self = { }

        streamer:setBounds(Vector3(-4000,-4000,-4000),Vector3(4000,4000,4000))
        self.world = streamer.physics
        self.world.stepSize = 1/100
        self.world.maxSteps = 20
        
        self.frameTimer = Timer()
        self.leftOver = 0

        self.speed = 1
        self.oneToOne = false
        self.enabled = true

        self.allocs = 0


        local function frameCallback()

                local count, allocs, reallocs, frees = get_alloc_stats()
                reset_alloc_stats()

                local iterations

                if self.enabled then

                        if self.oneToOne then
                                iterations = self.world:pump(self.world.stepSize)
                                self.frameTimer:reset()
                        else
                                local elapsed = self.leftOver +
                                        self.frameTimer.us/1000000 * self.speed

                                if elapsed > self.world.stepSize then
                                        self.frameTimer:reset()
                                        iterations = self.world:pump(elapsed)
                                end

                                self.leftOver =
                                        math.mod(elapsed,self.world.stepSize)
                        end

                else
                        self.frameTimer:reset()
                end

                if iterations then
                        local _, p_allocs = get_alloc_stats()
                        self.allocs = p_allocs / iterations
                else
                        self.allocs = 0
                end
                set_alloc_stats(allocs,reallocs,frees)

        end

        main.frameCallbacks:insert("Physics",frameCallback)

        ui:bind("F9", function ()
                self.enabled = not self.enabled
                echo("Physics enabled: "..tostring(self.enabled))
        end)
        ui:bind("F10", function ()
                if ui:ctrl() then
                        self.world.solverSplitImpulse = not self.world.solverSplitImpulse
                        echo("Physics splitImpulse: "..tostring(self.world.splitImpulse))
                else
                        self.oneToOne = not self.oneToOne
                        echo("Physics one-to-one: "..tostring(self.oneToOne))
                end
        end)
        ui:bind("F11", function ()
                if ui:ctrl() then
                        -- for recording movies
                        self.world:pump(1/25)
                        capturer:singleScreenShot()
                else
                        self.world:pump(self.world.stepSize)
                end
        end)

        make_instance(self,Physics)

        return self

end

function Physics:destroy()
        main.frameCallbacks:removeByName("Physics")
end

if physics ~= nil then
        physics:destroy()
end
physics = Physics.new()

pw = physics.world

-- vim: shiftwidth=8:tabstop=8:expandtab
